from builtins import object
import decimal
import datetime
import uuid

from enum import Enum
from past.types import basestring
from sqlalchemy import Column, VARCHAR, INTEGER, REAL, BIGINT, types, event
from sqlalchemy.orm import relation, foreign, object_session, backref

from ._common import hybrid_property_gncnumeric
from ._common import CallableList
from .sa_extra import _DateTime, DeclarativeBase, _Date


class KVP_Type(Enum):
    KVP_TYPE_INVALID = -1
    KVP_TYPE_GINT64 = 1
    KVP_TYPE_DOUBLE = 2
    KVP_TYPE_NUMERIC = 3
    KVP_TYPE_STRING = 4
    KVP_TYPE_GUID = 5
    KVP_TYPE_TIMESPEC = 6
    KVP_TYPE_BINARY = 7
    KVP_TYPE_GLIST = 8
    KVP_TYPE_FRAME = 9
    KVP_TYPE_GDATE = 10


pytype_KVPtype = {
    int: KVP_Type.KVP_TYPE_GINT64,
    float: KVP_Type.KVP_TYPE_DOUBLE,
    decimal.Decimal: KVP_Type.KVP_TYPE_NUMERIC,
    dict: KVP_Type.KVP_TYPE_FRAME,
    # to fill
}

KVPtype_fields = {
    KVP_Type.KVP_TYPE_GINT64: 'int64_val',
    KVP_Type.KVP_TYPE_DOUBLE: 'double_val',
    KVP_Type.KVP_TYPE_STRING: 'string_val',
    KVP_Type.KVP_TYPE_GUID: 'guid_val',
    KVP_Type.KVP_TYPE_TIMESPEC: 'timespec_val',
    KVP_Type.KVP_TYPE_GDATE: 'gdate_val',
    KVP_Type.KVP_TYPE_NUMERIC: ('numeric_val_num', 'numeric_val_denom'),
    KVP_Type.KVP_TYPE_FRAME: 'guid',
}


class SlotType(types.TypeDecorator):
    """Used to customise the DateTime type for sqlite (ie without the separators as in gnucash
    """
    impl = INTEGER

    def process_bind_param(self, value, dialect):
        if value is not None:
            return value.value

    def process_result_value(self, value, dialect):
        if value is not None:
            return KVP_Type(value)


class DictWrapper(object):
    def __contains__(self, key):
        for sl in self.slot_collection:
            if sl.name == key:
                return True
        else:
            return False


    def __getitem__(self, key):
        keys = key.split("/", 1)
        key = keys[0]
        for sl in self.slot_collection:
            if sl.name == key:
                break
        else:
            raise KeyError("No slot exists with name '{}'".format(key))
        if len(keys)>1:
            return sl[keys[1]]
        else:
            return sl #.value

    def __setitem__(self, key, value):
        keys = key.split("/", 1)
        key = keys[0]
        for sl in self.slot_collection:
            if sl.name == key:
                break
        else:
            # new key
            if len(keys)>1:
                if isinstance(self, SlotFrame):
                    sf = SlotFrame(name= self._name + "/" + key)
                else:
                    sf = SlotFrame(name= key)
                sf[keys[1]] = value
                self.slot_collection.append(sf)
            else:
                self.slot_collection.append(slot(parent=self, name=key, value=value))

            return
        if len(keys)>1:
            sl[keys[1]] = value
            return
        # assign if type is correct
        if isinstance(value, sl._python_type):
            sl.value = value
        else:
            raise TypeError("Type of '{}' is not one of {}".format(value, sl._python_type))

    def __delitem__(self, key):
        if isinstance(key,slice):
            # delete all
            del self.slot_collection[key]
            return
        keys = key.split("/", 1)
        for i, sl in enumerate(self.slot_collection):
            if sl.name == keys[0]:
                break
        else:
            raise KeyError("No slot exists with name '{}'".format(key))
        if len(keys)>1:
            del sl[keys[1]]
        else:
            del self.slot_collection[i]


    def iteritems(self):
        for sl in self.slot_collection:
            yield sl.name, sl

    def get(self, key, default=None):
        try:
            return self[key].value
        except KeyError:
            return default


class Slot(DeclarativeBase):
    __tablename__ = 'slots'

    # column definitions
    _name = Column('name', VARCHAR(length=4096), nullable=False)
    @property
    def name(self):
        if self._name:
            return self._name.split("/")[-1]
        else:
            return self._name

    @name.setter
    def name(self, value):
        self._name = value

    id = Column('id', INTEGER(), primary_key=True, nullable=False)
    obj_guid = Column('obj_guid', VARCHAR(length=32), nullable=False, index=True)
    slot_type = Column('slot_type', SlotType(), nullable=False)

    __mapper_args__ = {
        'polymorphic_on': slot_type,
    }

    def __repr__(self):
        return "<{} {}={}>".format(self.__class__.__name__, self.name, self.value)


class SlotSimple(Slot):
    __mapper_args__ = {
        'polymorphic_identity': -1,
    }

    _python_type = ()

    @property
    def value(self):
        return getattr(self, self._field)

    @value.setter
    def value(self, value):
        setattr(self, self._field, value)


def define_simpleslot(postfix, pytype, KVPtype, field, col_type, col_default):
    cls = type(
        'Slot{}'.format(postfix),
        (SlotSimple,),
        {
            "__mapper_args__": {'polymorphic_identity': KVPtype},
            field: Column(field, col_type, default=col_default),
            "_field": field,
            "_python_type": pytype,
        }
    )
    return cls


SlotInt = define_simpleslot(postfix="Int",
                            pytype=(int,),
                            KVPtype=KVP_Type.KVP_TYPE_GINT64,
                            field="int64_val",
                            col_type=BIGINT(),
                            col_default=0,
)
SlotDouble = define_simpleslot(postfix="Double",
                               pytype=(float,),
                               KVPtype=KVP_Type.KVP_TYPE_DOUBLE,
                               field="double_val",
                               col_type=REAL(),
                               col_default=0,
)
SlotTime = define_simpleslot(postfix="Time",
                             pytype=(datetime.time,),
                             KVPtype=KVP_Type.KVP_TYPE_TIMESPEC,
                             field="timespec_val",
                             col_type=_DateTime(),
                             col_default=None,
)
SlotDate = define_simpleslot(postfix="Date",
                             pytype=(datetime.date,),
                             KVPtype=KVP_Type.KVP_TYPE_GDATE,
                             field="gdate_val",
                             col_type=_Date(),
                             col_default=None,
)
SlotString = define_simpleslot(postfix="String",
                               pytype=(basestring,),
                               KVPtype=KVP_Type.KVP_TYPE_STRING,
                               field="string_val",
                               col_type=VARCHAR(length=4096),
                               col_default=None,
)


class SlotNumeric(Slot):
    __mapper_args__ = {
        'polymorphic_identity': KVP_Type.KVP_TYPE_NUMERIC
    }
    _python_type = (tuple, decimal.Decimal)

    _numeric_val_denom = Column('numeric_val_denom', BIGINT(), nullable=False, default=1)
    _numeric_val_num = Column('numeric_val_num', BIGINT(), nullable=False, default=0)
    value = hybrid_property_gncnumeric(_numeric_val_num, _numeric_val_denom)


class SlotFrame(DictWrapper, Slot):
    __mapper_args__ = {
        'polymorphic_identity': KVP_Type.KVP_TYPE_FRAME
    }
    _python_type = (dict,)

    guid_val = Column('guid_val', VARCHAR(length=32))

    slot_collection = relation('Slot',
                               primaryjoin=foreign(Slot.obj_guid) == guid_val,
                               cascade='all, delete-orphan',
                               collection_class=CallableList,
                               single_parent=True,
                               backref=backref("parent", remote_side=guid_val),
    )

    @property
    def value(self):
        # convert to dict
        return {sl.name: sl.value for sl in self.slot_collection}

    @value.setter
    def value(self, value):
        self.slot_collection = [slot(parent=self, name=k, value=v) for k, v in value.items()]


    def __init__(self, **kwargs):
        self.guid_val = uuid.uuid4().hex
        super(SlotFrame, self).__init__(**kwargs)


@event.listens_for(SlotFrame.slot_collection, 'remove')
def remove_slot(target, value, initiator):
    s = object_session(value)
    if value in s.new:
        s.expunge(value)
    else:
        s.delete(value)


def get_all_subclasses(cls):
    all_subclasses = []

    direct_subclasses = cls.__subclasses__()

    all_subclasses.extend(direct_subclasses)

    for subclass in direct_subclasses:
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses


def slot(parent, name, value):
    if isinstance(parent, SlotFrame):
        name = parent._name + "/" + name

    # handle datetime before others (as otherwise can be mixed with date)
    if isinstance(value, datetime.datetime):
        return SlotTime(name=name, value=value)

    for cls in get_all_subclasses(Slot):
        if isinstance(value, cls._python_type) and cls!=SlotFrame:
            return cls(name=name, value=value)

    if isinstance(value, dict):
        # transform a dict to Frame/Slots
        sf = SlotFrame(name=name)
        for k, v in value.items():
            sl = slot(parent=sf, name=k, value=v)
            # sl.obj_guid = sf.guid_val
            sl.parent = sf
        return sf

    raise ValueError("Cannot handle type of '{}'".format(value))



from __future__ import print_function
from piecash import create_book

# create by default an in memory sqlite version
ses = create_book(echo=False)
book = ses.book

print("Book is saved:", ses.is_saved, end=' ')
print(" ==> book description:", book.root_account.description)

print("changing description...")
book.root_account.description = "hello, book"
print("Book is saved:", ses.is_saved, end=' ')
print(" ==> book description:", book.root_account.description)

print("saving...")
ses.save()

print("Book is saved:", ses.is_saved, end=' ')
print(" ==> book description:", book.root_account.description)

print("changing description...")
book.root_account.description = "nevermind, book"
print("Book is saved:", ses.is_saved, end=' ')
print(" ==> book description:", book.root_account.description)

print("cancel...")
ses.cancel()

print("Book is saved:", ses.is_saved, end=' ')
print(" ==> book description:", book.root_account.description)



from sqlalchemy import Column, INTEGER, BIGINT, VARCHAR, ForeignKey
from sqlalchemy.orm import composite, relation, backref

# change of the __doc__ string as getting error in sphinx ==> should be reported to SA project
composite.__doc__ = None #composite.__doc__.replace(":ref:`mapper_composite`", "")

from .sa_extra import _Date, _DateTime, DeclarativeBase
from ._common import Address,CallableList
from ._declbase import DeclarativeBaseGuid


class Billterm(DeclarativeBaseGuid):
    __tablename__ = 'billterms'

    __table_args__ = {}

    # column definitions
    cutoff = Column('cutoff', INTEGER())
    description = Column('description', VARCHAR(length=2048), nullable=False)
    discount_denom = Column('discount_denom', BIGINT())
    discount_num = Column('discount_num', BIGINT())
    discountdays = Column('discountdays', INTEGER())
    duedays = Column('duedays', INTEGER())
    invisible = Column('invisible', INTEGER(), nullable=False)
    name = Column('name', VARCHAR(length=2048), nullable=False)
    parent = Column('parent', VARCHAR(length=32))
    refcount = Column('refcount', INTEGER(), nullable=False)
    type = Column('type', VARCHAR(length=2048), nullable=False)

    # relation definitions


class Customer(DeclarativeBaseGuid):
    __tablename__ = 'customers'

    __table_args__ = {}

    # column definitions
    active = Column('active', INTEGER(), nullable=False)
    addr_addr1 = Column('addr_addr1', VARCHAR(length=1024))
    addr_addr2 = Column('addr_addr2', VARCHAR(length=1024))
    addr_addr3 = Column('addr_addr3', VARCHAR(length=1024))
    addr_addr4 = Column('addr_addr4', VARCHAR(length=1024))
    addr_email = Column('addr_email', VARCHAR(length=256))
    addr_fax = Column('addr_fax', VARCHAR(length=128))
    addr_name = Column('addr_name', VARCHAR(length=1024))
    addr_phone = Column('addr_phone', VARCHAR(length=128))
    addr = composite(Address, addr_addr1, addr_addr2, addr_addr3, addr_addr4,
                     addr_email, addr_fax, addr_name, addr_phone)
    credit_denom = Column('credit_denom', BIGINT(), nullable=False)
    credit_num = Column('credit_num', BIGINT(), nullable=False)
    currency = Column('currency', VARCHAR(length=32), nullable=False)
    discount_denom = Column('discount_denom', BIGINT(), nullable=False)
    discount_num = Column('discount_num', BIGINT(), nullable=False)
    # guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False)
    id = Column('id', VARCHAR(length=2048), nullable=False)
    name = Column('name', VARCHAR(length=2048), nullable=False)
    notes = Column('notes', VARCHAR(length=2048), nullable=False)
    shipaddr_addr1 = Column('shipaddr_addr1', VARCHAR(length=1024))
    shipaddr_addr2 = Column('shipaddr_addr2', VARCHAR(length=1024))
    shipaddr_addr3 = Column('shipaddr_addr3', VARCHAR(length=1024))
    shipaddr_addr4 = Column('shipaddr_addr4', VARCHAR(length=1024))
    shipaddr_email = Column('shipaddr_email', VARCHAR(length=256))
    shipaddr_fax = Column('shipaddr_fax', VARCHAR(length=128))
    shipaddr_name = Column('shipaddr_name', VARCHAR(length=1024))
    shipaddr_phone = Column('shipaddr_phone', VARCHAR(length=128))
    shipaddr = composite(Address, shipaddr_addr1, shipaddr_addr2, shipaddr_addr3, shipaddr_addr4,
                         shipaddr_email, shipaddr_fax, shipaddr_name, shipaddr_phone)

    tax_included = Column('tax_included', INTEGER())
    tax_override = Column('tax_override', INTEGER(), nullable=False)
    taxtable = Column('taxtable', VARCHAR(length=32))
    terms = Column('terms', VARCHAR(length=32))

    # relation definitions


class Employee(DeclarativeBaseGuid):
    __tablename__ = 'employees'

    __table_args__ = {}

    # column definitions
    acl = Column('acl', VARCHAR(length=2048), nullable=False)
    active = Column('active', INTEGER(), nullable=False)
    addr_addr1 = Column('addr_addr1', VARCHAR(length=1024))
    addr_addr2 = Column('addr_addr2', VARCHAR(length=1024))
    addr_addr3 = Column('addr_addr3', VARCHAR(length=1024))
    addr_addr4 = Column('addr_addr4', VARCHAR(length=1024))
    addr_email = Column('addr_email', VARCHAR(length=256))
    addr_fax = Column('addr_fax', VARCHAR(length=128))
    addr_name = Column('addr_name', VARCHAR(length=1024))
    addr_phone = Column('addr_phone', VARCHAR(length=128))
    addr = composite(Address, addr_addr1, addr_addr2, addr_addr3, addr_addr4,
                     addr_email, addr_fax, addr_name, addr_phone)
    ccard_guid = Column('ccard_guid', VARCHAR(length=32))
    currency = Column('currency', VARCHAR(length=32), nullable=False)
    # guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False)
    id = Column('id', VARCHAR(length=2048), nullable=False)
    language = Column('language', VARCHAR(length=2048), nullable=False)
    rate_denom = Column('rate_denom', BIGINT(), nullable=False)
    rate_num = Column('rate_num', BIGINT(), nullable=False)
    username = Column('username', VARCHAR(length=2048), nullable=False)
    workday_denom = Column('workday_denom', BIGINT(), nullable=False)
    workday_num = Column('workday_num', BIGINT(), nullable=False)

    # relation definitions


class Entry(DeclarativeBaseGuid):
    __tablename__ = 'entries'

    __table_args__ = {}

    # column definitions
    action = Column('action', VARCHAR(length=2048))
    b_acct = Column('b_acct', VARCHAR(length=32))
    b_paytype = Column('b_paytype', INTEGER())
    b_price_denom = Column('b_price_denom', BIGINT())
    b_price_num = Column('b_price_num', BIGINT())
    b_taxable = Column('b_taxable', INTEGER())
    b_taxincluded = Column('b_taxincluded', INTEGER())
    b_taxtable = Column('b_taxtable', VARCHAR(length=32))
    bill = Column('bill', VARCHAR(length=32))
    billable = Column('billable', INTEGER())
    billto_guid = Column('billto_guid', VARCHAR(length=32))
    billto_type = Column('billto_type', INTEGER())
    date = Column('date', _DateTime(), nullable=False)
    date_entered = Column('date_entered', _DateTime())
    description = Column('description', VARCHAR(length=2048))
    # guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False)
    i_acct = Column('i_acct', VARCHAR(length=32))
    i_disc_how = Column('i_disc_how', VARCHAR(length=2048))
    i_disc_type = Column('i_disc_type', VARCHAR(length=2048))
    i_discount_denom = Column('i_discount_denom', BIGINT())
    i_discount_num = Column('i_discount_num', BIGINT())
    i_price_denom = Column('i_price_denom', BIGINT())
    i_price_num = Column('i_price_num', BIGINT())
    i_taxable = Column('i_taxable', INTEGER())
    i_taxincluded = Column('i_taxincluded', INTEGER())
    i_taxtable = Column('i_taxtable', VARCHAR(length=32))
    invoice = Column('invoice', VARCHAR(length=32))
    notes = Column('notes', VARCHAR(length=2048))
    order_guid = Column('order_guid', VARCHAR(length=32))
    quantity_denom = Column('quantity_denom', BIGINT())
    quantity_num = Column('quantity_num', BIGINT())

    # relation definitions


class Invoice(DeclarativeBaseGuid):
    __tablename__ = 'invoices'

    __table_args__ = {}

    # column definitions
    active = Column('active', INTEGER(), nullable=False)
    billing_id = Column('billing_id', VARCHAR(length=2048))
    billto_guid = Column('billto_guid', VARCHAR(length=32))
    billto_type = Column('billto_type', INTEGER())
    charge_amt_denom = Column('charge_amt_denom', BIGINT())
    charge_amt_num = Column('charge_amt_num', BIGINT())
    currency = Column('currency', VARCHAR(length=32), nullable=False)
    date_opened = Column('date_opened', _DateTime())
    date_posted = Column('date_posted', _DateTime())
    # guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False)
    id = Column('id', VARCHAR(length=2048), nullable=False)
    notes = Column('notes', VARCHAR(length=2048), nullable=False)
    owner_guid = Column('owner_guid', VARCHAR(length=32))
    owner_type = Column('owner_type', INTEGER())
    post_acc = Column('post_acc', VARCHAR(length=32))
    post_lot = Column('post_lot', VARCHAR(length=32))
    post_txn = Column('post_txn', VARCHAR(length=32))
    terms = Column('terms', VARCHAR(length=32))

    # relation definitions


class Job(DeclarativeBaseGuid):
    __tablename__ = 'jobs'

    __table_args__ = {}

    # column definitions
    active = Column('active', INTEGER(), nullable=False)
    # guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False)
    id = Column('id', VARCHAR(length=2048), nullable=False)
    name = Column('name', VARCHAR(length=2048), nullable=False)
    owner_guid = Column('owner_guid', VARCHAR(length=32))
    owner_type = Column('owner_type', INTEGER())
    reference = Column('reference', VARCHAR(length=2048), nullable=False)

    # relation definitions


class Order(DeclarativeBaseGuid):
    __tablename__ = 'orders'

    __table_args__ = {}

    # column definitions
    active = Column('active', INTEGER(), nullable=False)
    date_closed = Column('date_closed', _DateTime(), nullable=False)
    date_opened = Column('date_opened', _DateTime(), nullable=False)
    # guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False)
    id = Column('id', VARCHAR(length=2048), nullable=False)
    notes = Column('notes', VARCHAR(length=2048), nullable=False)
    owner_guid = Column('owner_guid', VARCHAR(length=32), nullable=False)
    owner_type = Column('owner_type', INTEGER(), nullable=False)
    reference = Column('reference', VARCHAR(length=2048), nullable=False)

    # relation definitions


class Vendor(DeclarativeBaseGuid):
    __tablename__ = 'vendors'

    __table_args__ = {}

    # column definitions
    active = Column('active', INTEGER(), nullable=False)
    addr_addr1 = Column('addr_addr1', VARCHAR(length=1024))
    addr_addr2 = Column('addr_addr2', VARCHAR(length=1024))
    addr_addr3 = Column('addr_addr3', VARCHAR(length=1024))
    addr_addr4 = Column('addr_addr4', VARCHAR(length=1024))
    addr_email = Column('addr_email', VARCHAR(length=256))
    addr_fax = Column('addr_fax', VARCHAR(length=128))
    addr_name = Column('addr_name', VARCHAR(length=1024))
    addr_phone = Column('addr_phone', VARCHAR(length=128))
    addr = composite(Address, addr_addr1, addr_addr2, addr_addr3, addr_addr4,
                     addr_email, addr_fax, addr_name, addr_phone)
    currency = Column('currency', VARCHAR(length=32), nullable=False)
    # guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False)
    id = Column('id', VARCHAR(length=2048), nullable=False)
    name = Column('name', VARCHAR(length=2048), nullable=False)
    notes = Column('notes', VARCHAR(length=2048), nullable=False)
    tax_inc = Column('tax_inc', VARCHAR(length=2048))
    tax_override = Column('tax_override', INTEGER(), nullable=False)
    tax_table = Column('tax_table', VARCHAR(length=32))
    terms = Column('terms', VARCHAR(length=32))

    # relation definitions


class Taxtable(DeclarativeBaseGuid):
    __tablename__ = 'taxtables'

    __table_args__ = {}

    # column definitions
    # guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False)
    invisible = Column('invisible', INTEGER(), nullable=False)
    name = Column('name', VARCHAR(length=50), nullable=False)
    parent = Column('parent', VARCHAR(length=32))
    refcount = Column('refcount', BIGINT(), nullable=False)

    # relation definitions


class TaxtableEntry(DeclarativeBase):
    __tablename__ = 'taxtable_entries'

    __table_args__ = {}

    # column definitions
    account = Column('account', VARCHAR(length=32), nullable=False)
    amount_denom = Column('amount_denom', BIGINT(), nullable=False)
    amount_num = Column('amount_num', BIGINT(), nullable=False)
    id = Column('id', INTEGER(), primary_key=True, nullable=False)
    taxtable = Column('taxtable', VARCHAR(length=32), nullable=False)
    type = Column('type', INTEGER(), nullable=False)

    # relation definitions


class Schedxaction(DeclarativeBaseGuid):
    __tablename__ = 'schedxactions'

    __table_args__ = {}

    # column definitions
    adv_creation = Column('adv_creation', INTEGER(), nullable=False)
    adv_notify = Column('adv_notify', INTEGER(), nullable=False)
    auto_create = Column('auto_create', INTEGER(), nullable=False)
    auto_notify = Column('auto_notify', INTEGER(), nullable=False)
    enabled = Column('enabled', INTEGER(), nullable=False)
    end_date = Column('end_date', _Date())
    instance_count = Column('instance_count', INTEGER(), nullable=False)
    last_occur = Column('last_occur', _Date())
    name = Column('name', VARCHAR(length=2048))
    num_occur = Column('num_occur', INTEGER(), nullable=False)
    rem_occur = Column('rem_occur', INTEGER(), nullable=False)
    start_date = Column('start_date', _Date())
    template_act_guid = Column('template_act_guid', VARCHAR(length=32), ForeignKey('accounts.guid'), nullable=False)

    # relation definitions


class Lot(DeclarativeBaseGuid):
    __tablename__ = 'lots'

    __table_args__ = {}

    # column definitions
    account_guid = Column('account_guid', VARCHAR(length=32), ForeignKey('accounts.guid'))
    is_closed = Column('is_closed', INTEGER(), nullable=False)

    # relation definitions
    account = relation('Account',
                       backref=backref("lots",
                                       cascade='all, delete-orphan',
                                       collection_class=CallableList, ))


import uuid

from sqlalchemy import Column, VARCHAR, ForeignKey, INTEGER
from sqlalchemy.orm import relation, backref, validates

from .._declbase import DeclarativeBaseGuid
from .._common import CallableList
from ..sa_extra import mapped_to_slot_property


root_types = {"ROOT"}
asset_types = {'RECEIVABLE', 'MUTUAL', 'CASH', 'ASSET', 'BANK', 'STOCK'}
liability_types = { 'CREDIT',  'LIABILITY', 'PAYABLE'}
income_types = {"INCOME"}
expense_types = {"EXPENSE"}
trading_types = {'TRADING'}
equity_types = {"EQUITY"}
#: the different types of accounts
ACCOUNT_TYPES = equity_types | income_types | expense_types | asset_types | liability_types | root_types | trading_types

# types that are compatible with other types
incexp_types = income_types | expense_types
assetliab_types = asset_types | liability_types


# types according to the sign of their balance
positive_types = asset_types | expense_types | trading_types
negative_types = liability_types | income_types | equity_types


def _is_parent_child_types_consistent(type_parent, type_child):
    """
    Return True if the child account is consistent with the parent account in terms of types, i.e.:

    1) if the parent is a root account, child can be anything but a root account
    2) if the child is a root account, it must have no parent account
    3) both parent and child are of the same family (asset, equity, income&expense, trading)

    Arguments
        type_parent(str): the type of the parent account
        type_child(str):  the type of the child account

    Returns
        True if both accounts are consistent, False otherwise
    """
    if type_parent in root_types:
        return type_child in (ACCOUNT_TYPES - root_types)

    if type_child in root_types:
        return type_parent is None

    for acc_types in (assetliab_types, equity_types, incexp_types, trading_types):
        if (type_child in acc_types) and (type_parent in acc_types):
            return True

    return False


class Account(DeclarativeBaseGuid):
    """
    A GnuCash Account which is specified by its name, type and commodity.

    Attributes:
        type (str): type of the Account
        code (str): code of the Account
        commodity (:class:`piecash.core.commodity.Commodity`): the commodity of the account
        commodity_scu (int): smallest currency unit for the account
        non_std_scu (int): 1 if the scu of the account is NOT the same as the commodity
        description (str): description of the account
        name (str): name of the account
        fullname (str): full name of the account (including name of parent accounts separated by ':')
        placeholder (int): 1 if the account is a placeholder (should not be involved in transactions)
        hidden (int): 1 if the account is hidden
        parent (:class:`Account`): the parent account of the account (None for the root account of a book)
        children (list of :class:`Account`): the list of the children accounts
    """
    __tablename__ = 'accounts'

    __table_args__ = {}

    # column definitions
    guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False, default=lambda: uuid.uuid4().hex)
    type = Column('account_type', VARCHAR(length=2048), nullable=False)
    code = Column('code', VARCHAR(length=2048))
    commodity_guid = Column('commodity_guid', VARCHAR(length=32), ForeignKey('commodities.guid'))
    _commodity_scu = Column('commodity_scu', INTEGER(), nullable=False)
    _non_std_scu = Column('non_std_scu', INTEGER(), nullable=False)

    @property
    def non_std_scu(self):
        return self._non_std_scu

    @property
    def commodity_scu(self):
        return self._commodity_scu

    @commodity_scu.setter
    def commodity_scu(self, value):
        if value is None:
            self._non_std_scu = 0
            if self.commodity:
                value = self.commodity.fraction
            else:
                value = 0
        else:
            self._non_std_scu = 1

        self._commodity_scu = value

    description = Column('description', VARCHAR(length=2048))
    hidden = Column('hidden', INTEGER())
    name = Column('name', VARCHAR(length=2048), nullable=False)

    parent_guid = Column('parent_guid', VARCHAR(length=32), ForeignKey('accounts.guid'))
    _placeholder = Column('placeholder', INTEGER())
    placeholder = mapped_to_slot_property(_placeholder, slot_name="placeholder",
                                          slot_transform=lambda v: "true" if v else None)

    # relation definitions
    commodity = relation('Commodity',
                         # single_parent=True,
                         backref=backref('accounts',
                                         cascade='all, delete-orphan',
                                         collection_class=CallableList))
    children = relation('Account',
                        backref=backref('parent', remote_side=guid),
                        cascade='all, delete-orphan',
                        collection_class=CallableList,
    )


    def __init__(self,
                 name,
                 type,
                 commodity,
                 parent=None,
                 description=None,
                 commodity_scu=None,
                 hidden=0,
                 placeholder=0,
                 code=None):
        self.name = name
        self.commodity = commodity
        self.type = type
        self.parent = parent
        self.description = description
        self.hidden = hidden
        self.placeholder = placeholder
        self.code = code
        self.commodity_scu = commodity_scu

    @validates('parent_guid', 'name')
    def validate_account_name(self, key, value):
        """
        Ensure the account name is unique amongst its sibling accounts
        """
        name = value if key == 'name' else self.name

        if self.parent:
            for acc in self.parent.children:
                if acc.name == name and acc != self:
                    raise ValueError("{} has two children with the same name {} : {} and {}".format(self.parent, name,
                                                                                                    acc, self))
        return value


    @validates('type', 'parent')
    def validate_type(self, key, value):
        """
        Ensure the account type is consistent
        """
        if key == "type":
            if value not in ACCOUNT_TYPES:
                raise ValueError("Account_type '{}' is not in {}".format(value, ACCOUNT_TYPES))

            if self.parent:
                if not _is_parent_child_types_consistent(self.parent.type, value):
                    raise ValueError("Child type '{}' is not consistent with parent type {}".format(
                        value, self.parent.type))

        if (key == "parent") and value and self.type:
            if not _is_parent_child_types_consistent(value.type, self.type):
                raise ValueError("Child type '{}' is not consistent with parent type {}".format(
                    self.type, value.type))

        return value

    @validates('commodity')
    def validate_commodity(self, key, value):
        """
        Ensure update of commodity_scu when commodity is changed
        """
        if value is None:
            return
        if self.commodity_scu is None or self.non_std_scu == 0:
            self.commodity_scu = value.fraction

        return value

    @property
    def fullname(self):
        if self.name:
            acc = self
            l = []
            while acc:
                l.append(acc.name)
                acc = acc.parent
            return ":".join(l[-2::-1])


    def get_balance(self):
        """

        Returns
            the balance of the account
        """
        return sum([sp.value for sp in self.splits]) * (-1 if self.type in negative_types else 1)


    def __repr__(self):
        if self.commodity:
            return "Account<{}[{}]>".format(self.fullname, self.commodity.mnemonic)
        else:
            return "Account<{}>".format(self.fullname)



from .session import create_book, open_book, Version
from .account import Account, ACCOUNT_TYPES
from .book import Book
from .commodity import Commodity, Price
from .transaction import Transaction, Split



import uuid

from sqlalchemy import Column, VARCHAR, event, inspect
from sqlalchemy.orm import relation, foreign, object_session

from .kvp import DictWrapper, Slot
from ._common import CallableList
from .sa_extra import DeclarativeBase


class DeclarativeBaseGuid(DictWrapper, DeclarativeBase):
    __abstract__ = True

    #: the unique identifier of the object
    guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False, default=lambda: uuid.uuid4().hex)

    # set the relation to the slots table (KVP)
    @classmethod
    def __declare_last__(cls):
        # do not do it on the DeclarativeBaseGuid as it is an abstract class
        if cls == DeclarativeBaseGuid:
            return

        cls.slots = relation('Slot',
                             primaryjoin=foreign(Slot.obj_guid) == cls.guid,
                             cascade='all, delete-orphan',
                             collection_class=CallableList,
        )

        # assign id of slot when associating to object
        @event.listens_for(cls.slots, "remove")
        def my_append_listener_slots(target, value, initiator):
            s = object_session(value)
            if s:
                if value in s.new:
                    s.expunge(value)
                else:
                    s.delete(value)


    def __init__(self, *args, **kwargs):
        """A simple constructor that allows initialization from kwargs.

        Sets attributes on the constructed instance using the names and
        values in ``kwargs``.

        Only keys that are present as
        attributes of the instance's class are allowed. These could be,
        for example, any mapped columns or relationships.
        """
        cls_ = type(self)
        key_rel = {rel.key: rel.mapper.class_ for rel in inspect(cls_).relationships}

        for k, v in kwargs.items():
            attr = getattr(cls_, k)
            # if the field is a relation and the value is a string, replace the string by the lookup
            if isinstance(v, str) and k in key_rel:
                v = key_rel[k].lookup(v)
            setattr(self, k, v)


    def get_session(self):
        # return the sa session of the object
        return object_session(self)

    @property
    def slot_collection(self):
        return self.slots


# -*- coding: utf-8 -*-
"""Project metadata

Information describing the project.
"""

# The package name, which is also the "UNIX name" for the project.
package = 'piecash'
project = "Python interface to GnuCash documents"
project_no_spaces = project.replace(' ', '')
version = '0.4.4'
description = 'A pythonic interface to GnuCash SQL documents.'
authors = ['sdementen']
authors_string = ', '.join(authors)
emails = ['sdementen@gmail.com']
license = 'MIT'
copyright = '2014 ' + authors_string
url = 'https://github.com/sdementen/piecash'



from decimal import Decimal

from sqlalchemy import Column, VARCHAR, INTEGER, cast, Float
from sqlalchemy.ext.hybrid import hybrid_property

from .sa_extra import DeclarativeBase, _Date, long


class GnucashException(Exception):
    pass


class GncNoActiveSession(GnucashException):
    pass


class GncValidationError(GnucashException):
    pass


class Recurrence(DeclarativeBase):
    __tablename__ = 'recurrences'

    __table_args__ = {}

    # column definitions
    id = Column('id', INTEGER(), primary_key=True, nullable=False)
    obj_guid = Column('obj_guid', VARCHAR(length=32), nullable=False)
    recurrence_mult = Column('recurrence_mult', INTEGER(), nullable=False)
    recurrence_period_start = Column('recurrence_period_start', _Date(), nullable=False)
    recurrence_period_type = Column('recurrence_period_type', VARCHAR(length=2048), nullable=False)
    recurrence_weekend_adjust = Column('recurrence_weekend_adjust', VARCHAR(length=2048), nullable=False)

    # relation definitions
    # added from the DeclarativeBaseGUID object (as linked from different objects like the slots)
    def __repr__(self):
        return "{}/{} from {} [{}]".format(self.recurrence_mult, self.recurrence_period_type,
                                           self.recurrence_period_start, self.recurrence_weekend_adjust)



class Address(object):
    _address_fields = "addr1 addr2 addr3 addr4 email fax name phone".split()

    def __init__(self, *args):
        for fld, val in zip(Address._address_fields, args):
            setattr(self, fld, val)

    def __composite_values__(self):
        return tuple(self)

    def __eq__(self, other):
        return isinstance(other, Address) and all(getattr(other, fld) == getattr(self, fld) for fld in Address._address_fields)

    def __ne__(self, other):
        return not self.__eq__(other)


def hybrid_property_gncnumeric(num_col, denom_col):
    """Return an hybrid_property handling a Decimal represented by a numerator and a denominator column.
    It assumes the python field related to the sqlcolumn is named as _sqlcolumn.

    :type num_col: sqlalchemy.sql.schema.Column
    :type denom_col: sqlalchemy.sql.schema.Column
    :return: sqlalchemy.ext.hybrid.hybrid_property
    """
    num_name, denom_name = "_{}".format(num_col.name), "_{}".format(denom_col.name)

    def fset(self, d):
        if d is None:
            num, denom = None, None
        else:
            if isinstance(d, tuple):
                d = Decimal(d[0]) / d[1]
            elif isinstance(d, (float, int, long, str)):
                d = Decimal(d)
            assert isinstance(d, Decimal)

            sign, digits, exp = d.as_tuple()
            denom = 10 ** max(-exp, 0)

            # print num_name, denom,
            denom_basis = getattr(self, "{}_basis".format(denom_name), None)
            if denom_basis is not None:
                # print "got a basis for ", self, denom_basis
                denom = denom_basis
            # print denom
            num = int(d * denom)

        setattr(self, num_name, num)
        setattr(self, denom_name, denom)


    def fget(self):
        num, denom = getattr(self, num_name), getattr(self, denom_name)
        if num:
            return Decimal(num) / denom


    def expr(cls):
        return cast(num_col, Float) / denom_col

    return hybrid_property(
        fget=fget,
        fset=fset,
        expr=expr,
    )


class CallableList(list):
    """
    A simple class (inherited from list) allowing to retrieve a given list element with a filter on an attribute.

    It can be used as the collection_class of a sqlalchemy relationship or to wrap any list (see examples
    in :class:`piecash.core.session.GncSession`)
    """
    def get(self, **kwargs):
        """
        Return the first element of the list that has attributes matching the kwargs dict.
        To be used as::

            l.get(mnemonic="EUR", namespace="CURRENCY")
        """
        for obj in self:
            for k, v in kwargs.items():
                if getattr(obj, k) != v:
                    break
            else:
                return obj
        else:
            raise KeyError("Could not find object with {} in {}".format(kwargs, self))


from collections import defaultdict
from decimal import Decimal
import datetime

from sqlalchemy import Column, VARCHAR, ForeignKey, BIGINT, event
from sqlalchemy.orm import relation, backref, validates
from sqlalchemy.orm.base import instance_state
from sqlalchemy.orm.exc import NoResultFound

from .._common import GncValidationError, hybrid_property_gncnumeric
from .._declbase import DeclarativeBaseGuid
from .._common import CallableList
from ..sa_extra import _DateTime, Session, mapped_to_slot_property
from .book import Book
from .account import Account


"""
Examples of transaction and splits (with value and quantity) for several transactions,
some mono-currency (in default or foreign currency), some multi-currency

Commodity<CURRENCY:EUR>    salary
    [Commodity<CURRENCY:EUR>] -1000 / -1000 for Account<Income>
    [Commodity<CURRENCY:EUR>] 1000 / 1000 for Account<Assets:Current Assets:Checking Account>
Commodity<CURRENCY:EUR>    transfert to US account
    [Commodity<CURRENCY:EUR>] -400 / -400 for Account<Assets:Current Assets:Checking Account>
    [Commodity<CURRENCY:USD>] 400 / 448.15 for Account<Assets:Current Assets:CheckAcc USD>
    [Commodity<CURRENCY:USD>] -400 / -448.15 for Account<Trading:CURRENCY:USD>
    [Commodity<CURRENCY:EUR>] 400 / 400 for Account<Trading:CURRENCY:EUR>
Commodity<CURRENCY:EUR>    other transfer + expense
    [Commodity<CURRENCY:EUR>] -210 / -210 for Account<Assets:Current Assets:Checking Account>
    [Commodity<CURRENCY:USD>] 182.85 / 213.21 for Account<Assets:Current Assets:CheckAcc USD>
    [Commodity<CURRENCY:USD>] -182.85 / -213.21 for Account<Trading:CURRENCY:USD>
    [Commodity<CURRENCY:EUR>] 182.85 / 182.85 for Account<Trading:CURRENCY:EUR>
    [Commodity<CURRENCY:EUR>] 17.15 / 17.15 for Account<Expenses>
    [Commodity<CURRENCY:EUR>] 10 / 10 for Account<Imbalance-EUR>
Commodity<CURRENCY:USD>    bonus
    [Commodity<CURRENCY:USD>] -150 / -150 for Account<Income:income in usd>
    [Commodity<CURRENCY:USD>] 150 / 150 for Account<Assets:Current Assets:CheckAcc USD>
Commodity<CURRENCY:USD>    retransfer
    [Commodity<CURRENCY:USD>] -100 / -100 for Account<Assets:Current Assets:CheckAcc USD>
    [Commodity<CURRENCY:EUR>] 100 / 90 for Account<Assets:Current Assets:Checking Account>
    [Commodity<CURRENCY:EUR>] -100 / -90 for Account<Trading:CURRENCY:EUR>
    [Commodity<CURRENCY:USD>] 100 / 100 for Account<Trading:CURRENCY:USD>
Commodity<CURRENCY:CAD>    cross CAD to USD transfer
    [Commodity<CURRENCY:CAD>] 30 / 30 for Account<Assets:Current Assets:CheckAcc CAD>
    [Commodity<CURRENCY:USD>] -30 / -26.27 for Account<Assets:Current Assets:CheckAcc USD>
    [Commodity<CURRENCY:USD>] 30 / 26.27 for Account<Trading:CURRENCY:USD>
    [Commodity<CURRENCY:CAD>] -30 / -30 for Account<Trading:CURRENCY:CAD>
Commodity<CURRENCY:USD>    cross CAD to USD transfer (initiated from USD account)
    [Commodity<CURRENCY:USD>] -26.27 / -26.27 for Account<Assets:Current Assets:CheckAcc USD>
    [Commodity<CURRENCY:CAD>] 26.27 / 30 for Account<Assets:Current Assets:CheckAcc CAD>
    [Commodity<CURRENCY:CAD>] -26.27 / -30 for Account<Trading:CURRENCY:CAD>
    [Commodity<CURRENCY:USD>] 26.27 / 26.27 for Account<Trading:CURRENCY:USD>
    """


class Split(DeclarativeBaseGuid):
    """
    A GnuCash Split.

    Attributes:
        transaction(:class:`piecash.core.transaction.Transaction`): transaction of the split
        account(:class:`piecash.core.account.Account`): account of the split
        lot(str): lot to which the split pertains
        memo(str): memo of the split
        value(:class:`decimal.Decimal`): amount express in the currency of the transaction of the split
        quantity(:class:`decimal.Decimal`): amount express in the commodity of the account of the split
        reconcile_state(str): 'n', 'c' or 'y'
        reconcile_date(:class:`datetime.datetime`): time
        action(str): new in GnuCash 2.6. usage not yet understood
    """
    __tablename__ = 'splits'

    __table_args__ = {}

    # column definitions
    account_guid = Column('account_guid', VARCHAR(length=32), ForeignKey('accounts.guid'), nullable=False, index=True)
    action = Column('action', VARCHAR(length=2048), nullable=False, default="")
    lot_guid = Column('lot_guid', VARCHAR(length=32), ForeignKey('lots.guid'))
    memo = Column('memo', VARCHAR(length=2048), nullable=False, default="")

    _quantity_denom = Column('quantity_denom', BIGINT(), nullable=False)
    _quantity_denom_basis = None
    _quantity_num = Column('quantity_num', BIGINT(), nullable=False)
    quantity = hybrid_property_gncnumeric(_quantity_num, _quantity_denom)

    reconcile_date = Column('reconcile_date', _DateTime())
    reconcile_state = Column('reconcile_state', VARCHAR(length=1), nullable=False, default="n")
    tx_guid = Column('tx_guid', VARCHAR(length=32), ForeignKey('transactions.guid'), nullable=False, index=True)

    _value_denom = Column('value_denom', BIGINT(), nullable=False)
    _value_num = Column('value_num', BIGINT(), nullable=False)
    _value_denom_basis = None
    value = hybrid_property_gncnumeric(_value_num, _value_denom)

    # relation definitions
    account = relation('Account', backref=backref('splits',
                                                  cascade='all, delete-orphan',
                                                  collection_class=CallableList,
    ))

    lot = relation('Lot', backref=backref('splits',
                                          cascade='all, delete-orphan',
                                          collection_class=CallableList,
    ))

    def __init__(self,
                 account=None,
                 value=0,
                 quantity=0,
                 transaction=None,
                 memo="",
                 reconcile_date=None,
                 reconcile_state="n",
                 lot=None,
    ):
        self.transaction = transaction
        self.account = account
        self.value = value
        self.quantity = quantity
        self.memo = memo
        self.reconcile_date=reconcile_date
        self.reconcile_state=reconcile_state
        self.lot=lot

    def __repr__(self):
        try:
            cur = self.transaction.currency.mnemonic
            acc = self.account
            com = acc.commodity.mnemonic
            if cur == com:
                return "<Split {} {} {}>".format(acc,
                                                 self.value, cur)
            else:
                return "<Split {} {} {} [{} {}]>".format(acc,
                                                         self.value, cur,
                                                         self.quantity, com)
        except AttributeError:
            return "<Split {}>".format(self.account)

    @validates("transaction", "account")
    def set_denom_basis(self, key, value):
        if value is None:
            return value
        if "transaction" == key:
            self._value_denom_basis = value.currency.fraction
            self.value = self.value
            trx = value
            acc = self.account
        if "account" == key:
            # if the account is already defined
            if self.account:
                # check that we keep the same commodity across the account change
                if self.account.commodity != value.commodity:
                    raise GncValidationError("The commodity of the new account of this split is not the same as the old account")
                if self.account.commodity_scu > value.commodity_scu:
                    raise GncValidationError("The commodity_scu of the new account of this split is lower than the one of the old account")

            self._quantity_denom_basis = value.commodity_scu
            self.quantity = self.quantity
            trx = self.transaction
            acc = value

        if trx and acc:
            if trx.currency == acc.commodity:
                self.quantity = self.value
                # if the quantity has different rounding that value, then reassign the quantity to the value
                if self.quantity != self.value:
                    self.value = self.quantity

        return value


class Transaction(DeclarativeBaseGuid):
    """
    A GnuCash Transaction.

    Attributes:
        currency (:class:`piecash.core.commodity.Commodity`): currency of the transaction. This attribute is
            write-once (i.e. one cannot change it after being set)
        description (str): description of the transaction
        enter_date (:class:`datetime.datetime`): time at which transaction is entered
        post_date (:class:`datetime.datetime`): day on which transaction is posted
        num (str): user provided transaction number
        splits (list of :class:`Split`): list of the splits of the transaction
    """
    __tablename__ = 'transactions'

    __table_args__ = {}

    # column definitions
    currency_guid = Column('currency_guid', VARCHAR(length=32), ForeignKey('commodities.guid'), nullable=False)
    description = Column('description', VARCHAR(length=2048))
    enter_date = Column('enter_date', _DateTime)
    num = Column('num', VARCHAR(length=2048), nullable=False)
    _post_date = Column('post_date', _DateTime, index=True)
    post_date = mapped_to_slot_property(_post_date, slot_name="date-posted",
                                        slot_transform=lambda x: x.date() if x else None)

    splits = relation(Split,
                      backref=backref('transaction'),
                      single_parent=True,
                      cascade='all, delete-orphan',
                      collection_class=CallableList,
    )


    # relation definitions
    currency = relation('Commodity', backref=backref('transactions',
                                                     cascade='all, delete-orphan',
                                                     collection_class=CallableList,
    ))

    def __init__(self,
                 currency,
                 description="",
                 splits=[],
                 enter_date=None,
                 post_date=None,
                 num="",
    ):
        self.currency = currency
        self.description = description
        self.enter_date = enter_date if enter_date else datetime.datetime.today()
        self.post_date = post_date if post_date else datetime.datetime.today().replace(hour=0, minute=0, second=0, microsecond=0)
        self.num = num
        self.splits = splits


    @validates("currency")
    def validate_currency(self, key, value):
        if value and value.namespace!="CURRENCY":
            raise GncValidationError("You are assigning a non currency commodity to a transaction")
        return value

    def validate(self, session):
        old = instance_state(self).committed_state

        # check same currency
        if "currency" in old and old["currency"] is not None:
            raise GncValidationError("You cannot change the currency of a transaction once it has been set")

        # validate the splits
        if "splits" in old:
            imbalance = Decimal(0)
            c = self.currency
            for sp in self.splits:
                if sp.account.commodity != c:
                    raise GncValidationError("Only single currency transactions are supported")

                sp.quantity = sp.value
                if sp.quantity != sp.value:
                    sp.value = sp.quantity

                imbalance += sp.value

            # if there is an imbalance, add an imbalance split to the transaction
            if imbalance:
                # retrieve imbalance account
                imb_acc_name = "Imbalance-{}".format(c.mnemonic)
                try:
                    acc = session.query(Account).filter_by(name=imb_acc_name).one()
                except NoResultFound:
                    book = session.query(Book).one()
                    acc = Account(name=imb_acc_name,
                                  parent=book.root_account,
                                  commodity=c,
                                  type="BANK")

                Split(value=-imbalance,
                      quantity=-imbalance,
                      account=acc,
                      transaction=self)


    @classmethod
    def single_transaction(cls,
                           post_date,
                           enter_date,
                           description,
                           value,
                           from_account,
                           to_account):
        # currency is derived from "from_account" (as in GUI)
        currency = from_account.commodity
        # currency of other destination account should be identical (as only one value given)
        assert currency == to_account.commodity, "Commodities of accounts should be the same"
        tx = Transaction(
            currency=currency,
            post_date=post_date,
            enter_date=enter_date,
            description=description,
            splits=[
                Split(account=from_account, value=-value),
                Split(account=to_account, value=value),
            ])
        return tx

    def __repr__(self):
        return "<Transaction in {} on {} ({})>".format(self.currency, self.post_date, self.enter_date)


@event.listens_for(Session, 'before_flush')
def set_imbalance_on_transaction(session, flush_context, instances):
    # identify transactions to verify
    txs = set()
    for o in session.dirty:
        if isinstance(o, Transaction):
            txs.add(o)
        if isinstance(o, Split):
            txs.add(o.transaction)
    txs = txs.union(o for o in session.new if isinstance(o, Transaction))

    # for each transaction, validate the transaction
    for tx in txs:
        tx.validate(session)



# -*- coding: utf-8 -*-
"""Python interface to GnuCash documents"""
from . import metadata

__version__ = metadata.version
__author__ = metadata.authors[0]
__license__ = metadata.license
__copyright__ = metadata.copyright

from ._common import GncNoActiveSession, GnucashException
from .core import (Book,
                   Account, ACCOUNT_TYPES,
                   Transaction, Split,
                   Commodity, Price,
                   create_book, open_book
)
from .business import Lot, Invoice, Employee, Customer, Job, Vendor  # must import as Transaction has a relation to it
from .budget import Budget, BudgetAmount
from .kvp import slot



from __future__ import division
import uuid

from sqlalchemy import Column, VARCHAR, INTEGER, BIGINT, ForeignKey
from sqlalchemy.orm import relation, backref, foreign

from .sa_extra import DeclarativeBase
from ._declbase import DeclarativeBaseGuid
from ._common import hybrid_property_gncnumeric, Recurrence, CallableList


class Budget(DeclarativeBaseGuid):
    __tablename__ = 'budgets'

    __table_args__ = {}

    # column definitions
    guid = Column('guid', VARCHAR(length=32), primary_key=True, nullable=False, default=lambda: uuid.uuid4().hex)
    description = Column('description', VARCHAR(length=2048))
    name = Column('name', VARCHAR(length=2048), nullable=False)
    num_periods = Column('num_periods', INTEGER(), nullable=False)

    # # relation definitions
    recurrence = relation(Recurrence,
                          primaryjoin=foreign(Recurrence.obj_guid) == guid,
                          cascade='all, delete-orphan',
                          uselist=False)

    def __repr__(self):
        return "<Budget {}({}) for {} periods following pattern '{}' >".format(self.name, self.description,
                                                                               self.num_periods, self.recurrence)


class BudgetAmount(DeclarativeBase):
    __tablename__ = 'budget_amounts'

    __table_args__ = {}

    # column definitions
    account_guid = Column('account_guid', VARCHAR(length=32),
                          ForeignKey('accounts.guid'), nullable=False)
    _amount_denom = Column('amount_denom', BIGINT(), nullable=False)
    _amount_num = Column('amount_num', BIGINT(), nullable=False)
    _amount_denom_basis = None
    amount = hybrid_property_gncnumeric(_amount_num, _amount_denom)

    budget_guid = Column('budget_guid', VARCHAR(length=32),
                         ForeignKey('budgets.guid'), nullable=False)
    id = Column('id', INTEGER(), primary_key=True, nullable=False)
    period_num = Column('period_num', INTEGER(), nullable=False)

    # relation definitions
    account = relation('Account', backref=backref('budget_amounts',
                                                  cascade='all, delete-orphan',
                                                  collection_class=CallableList, ))
    budget = relation('Budget', backref=backref('amounts',
                                                cascade='all, delete-orphan',
                                                collection_class=CallableList, ))

    def __repr__(self):
        return "<BudgetAmount {}={}>".format(self.period_num, self.amount)





#!/usr/bin/env python
# # @file
# @brief Creates a basic set of accounts and a couple of transactions
# @ingroup python_bindings_examples
from decimal import Decimal

from piecash import create_book, Account, Transaction, Split, Commodity

FILE_1 = "/tmp/example.gnucash"

with create_book(FILE_1, overwrite=True) as session:
    book = session.book
    root_acct = book.root_account
    cad = Commodity.create_currency_from_ISO("CAD")
    expenses_acct = Account(parent=root_acct,
                            name="Expenses",
                            type="EXPENSE",
                            commodity=cad)
    savings_acct = Account(parent=root_acct,
                           name="Savings",
                           type="BANK",
                           commodity=cad)
    opening_acct = Account(parent=root_acct,
                           name="Opening Balance",
                           type="EQUITY",
                           commodity=cad)
    num1 = Decimal("4")
    num2 = Decimal("100")
    num3 = Decimal("15")

    # create transaction with core objects in one step
    trans1 = Transaction(currency=cad,
                         description="Groceries",
                         splits=[
                             Split(value=num1, account=expenses_acct),
                             Split(value=-num1, account=savings_acct),
                         ])

    # create transaction with core object in multiple steps
    trans2 = Transaction(currency=cad,
                         description="Opening Savings Balance")

    split3 = Split(value=num2,
                   account=savings_acct,
                   transaction=trans2)

    split4 = Split(value=-num2,
                   account=opening_acct,
                   transaction=trans2)

    # create transaction with factory function
    trans3 = Transaction.single_transaction(None,None,"Pharmacy", num3, savings_acct, expenses_acct)

    session.save()





from sqlalchemy import Column, VARCHAR, ForeignKey
from sqlalchemy.orm import relation, backref

from .._declbase import DeclarativeBaseGuid


class Book(DeclarativeBaseGuid):
    """
    A Book represents an accounting book. A new GnuCash document contains only a single Book .
    """
    __tablename__ = 'books'

    __table_args__ = {}

    # column definitions
    root_account_guid = Column('root_account_guid', VARCHAR(length=32),
                               ForeignKey('accounts.guid'), nullable=False)
    root_template_guid = Column('root_template_guid', VARCHAR(length=32),
                                ForeignKey('accounts.guid'), nullable=False)

    # relation definitions
    #: the root account of the book
    root_account = relation('Account', foreign_keys=[root_account_guid],
                            backref=backref('book', cascade='all, delete-orphan', uselist=False))
    #: the root template account of the book (usage not yet clear...)
    root_template = relation('Account', foreign_keys=[root_template_guid])

    def __init__(self, root_account, root_template):
        self.root_account=root_account
        self.root_template=root_template




# -*- coding: utf-8 -*-
from builtins import object
import shutil
import os
import pytest

from piecash import Transaction, Commodity, open_book, create_book, Account
from piecash.kvp import Slot
from piecash._common import GnucashException
from piecash.core.session import Version, gnclock

from test_helper import file_template, file_for_test


@pytest.fixture
def session(request):
    s = create_book()
    return s

@pytest.fixture
def sa_session(request):
    s = create_book()
    return s.sa_session


@pytest.fixture
def session_readonly(request):
    shutil.copyfile(file_template, file_for_test)

    # default session is readonly
    s = open_book(file_for_test)

    request.addfinalizer(lambda: os.remove(file_for_test))
    return s

@pytest.fixture
def session_readonly_lock(request):
    shutil.copyfile(file_template, file_for_test)

    # default session is readonly
    s = open_book(file_for_test, acquire_lock=True)

    def close_s():
        s.close()
        os.remove(file_for_test)
    request.addfinalizer(close_s)
    return s


class TestModelCore_EmptyBook(object):
    def test_accounts(self, session):
        # two accounts in an empty gnucash file
        account_names = session.query(Account.name).all()

        assert set(account_names) == {(u'Template Root',),
                                      (u'Root Account',),
        }

    def test_transactions(self, session):
        # no transactions in an empty gnucash file
        transactions = session.query(Transaction).all()
        assert transactions == []

    def test_commodities(self, session):
        # no commodities in an empty gnucash file
        commodities = session.query(Commodity.mnemonic).all()
        assert commodities == [("EUR",)]

    def test_slots(self, session):
        # no slots  in an empty gnucash file
        slots = session.query(Slot).all()
        assert slots == []

    def test_versions(self, session):
        # confirm versions of tables
        versions = session.query(Version.table_name,
                                 Version.table_version).all()
        assert set(versions) == {(u'Gnucash', 2060400), (u'Gnucash-Resave', 19920),
                                 (u'accounts', 1), (u'books', 1),
                                 (u'budgets', 1), (u'budget_amounts', 1), ('jobs', 1), (u'orders', 1),
                                 (u'taxtables', 2), (u'taxtable_entries', 3), (u'vendors', 1), (u'recurrences', 2),
                                 (u'slots', 3), (u'transactions', 3), (u'splits', 4), (u'lots', 2), (u'entries', 3),
                                 (u'billterms', 2), (u'invoices', 3), (u'commodities', 1), (u'schedxactions', 1),
                                 (u'prices', 2), (u'customers', 2), (u'employees', 2),
        }

    def test_readonly_true(self, session_readonly):
        # control exception when adding object to readonly gnucash db
        v = Version(table_name="sample", table_version="other sample")
        sa_session_readonly = session_readonly.sa_session
        sa_session_readonly.add(v)
        with pytest.raises(GnucashException):
            sa_session_readonly.commit()

        # control exception when deleting object to readonly gnucash db
        sa_session_readonly.delete(session_readonly.query(Account).first())
        with pytest.raises(GnucashException):
            sa_session_readonly.commit()

        # control exception when modifying object to readonly gnucash db
        sa_session_readonly.query(Account).first().name = "foo"
        with pytest.raises(GnucashException):
            sa_session_readonly.commit()


    def test_readonly_false(self, sa_session):
        v = Version(table_name="fo", table_version="ok")
        sa_session.add(v)
        assert sa_session.flush() is None

    def test_lock(self, session_readonly_lock):
        # test that lock is not taken in readonly session
        locks = list(session_readonly_lock.sa_session.execute(gnclock.select()))
        assert len(locks) == 0


# class TestModelCore_CreateObjects(object):
#     def test_accounts(self, session):
#         # two accounts in an empty gnucash file
#         # Account(type=)
#         account_names = session.query(Account.name).all()
#
#         assert set(account_names) == {(u'Template Root',),
#                                       (u'Root Account',),
#         }
#
#     def test_transactions(self, session):
#         # no transactions in an empty gnucash file
#         transactions = session.query(Transaction).all()
#         assert transactions == []
#
#     def test_commodities(self, session):
#         # no commodities  in an empty gnucash file
#         commodities = session.query(Commodity).all()
#         assert commodities == []
#
#     def test_slots(self, session):
#         # no slots  in an empty gnucash file
#         slots = session.query(Slot).all()
#         assert slots == []
#
#     def test_versions(self, session):
#         # confirm versions of tables
#         versions = session.query(Version.table_name,
#                                  Version.table_version).all()
#         assert set(versions) == {(u'Gnucash', 2060400), (u'Gnucash-Resave', 19920),
#                                  (u'accounts', 1), (u'books', 1),
#                                  (u'budgets', 1), (u'budget_amounts', 1), ('jobs', 1), (u'orders', 1),
#                                  (u'taxtables', 2), (u'taxtable_entries', 3), (u'vendors', 1), (u'recurrences', 2),
#                                  (u'slots', 3), (u'transactions', 3), (u'splits', 4), (u'lots', 2), (u'entries', 3),
#                                  (u'billterms', 2), (u'invoices', 3), (u'commodities', 1), (u'schedxactions', 1),
#                                  (u'prices', 2), (u'customers', 2), (u'employees', 2),
#         }